/*
fuzzy.cc

This C++ file defines member functions for the class FUZZY_MODULE
It was generated with rcsdesign
with template version 1.1

MODIFICATIONS:
5-Oct-99	Modified with RCS-Design tool.
21-Sep-99	Created with RCS-Design tool.

*/

// Include Files
#include "rcs.hh" 	// Common RCS definitions
#include "nml_mod.hh" 	// NML_MODULE definitions
#include "fuzzy.hh" 	// FUZZY_MODULE definitions
// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.


#include "fuzzyn.hh" 	// NML Commands and Status definitions for fuzzy
// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


// Constructor
FUZZY_MODULE::FUZZY_MODULE(int _is_base_class)
{

	// Initialize the NML channels if this module is not being used as the base class for another module.
	if(!_is_base_class)
	{
		INITIALIZE_NML();
	}
	// Add additional code to initialize the module here.
	num_of_fuzzy_inputs = NUM_OF_FUZZY_INPUTS;
	num_of_fuzzy_outputs = NUM_OF_FUZZY_OUTPUTS;
	fuzzy_sys = new FUZZY_SYS(num_of_fuzzy_inputs, num_of_fuzzy_outputs);

	if(fuzzy_sys->fuzzy_sys_err != FUZZY_SYS_NO_ERROR)
	{
		logError("Error in creating FUZZY_MODULE object!!!");
		status = RCS_ERROR; // let my supervisor know I have a problem.
	}

}

// Overloaded Virtual Functions

/*
INITIALIZE_NML

The INITIALIZE_NML function is ussually called only once from within the
constructor. It should not be called if a derived class will also call it.

*/
void FUZZY_MODULE::INITIALIZE_NML()
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	setErrorLogChannel(new NML(nmlErrorFormat, "errlog", "fuzzy", "fuzzy_module.nml"));
	setCmdChannel(new RCS_CMD_CHANNEL(fuzzyFormat, "fuzzy_cmd", "fuzzy", "fuzzy_module.nml"));
	fuzzy_status = new FUZZY_STATUS();
	setStatChannel(new RCS_STAT_CHANNEL(fuzzyFormat, "fuzzy_sts", "fuzzy", "fuzzy_module.nml"), fuzzy_status);
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


}

/*
PRE_PROCESS

The PRE_PROCESS function is called every cycle after the command and
subordinates status have been read but before DECISION_PROCESS is called.
It is intended to be used for tasks such as sensory processing that should
be performed every cycle regardless of the current command or state.

*/
void FUZZY_MODULE::PRE_PROCESS()
{
	// Pre-Processing Code
	// Read and/or calculate inputs to the fuzzy system 
	for(int ii=0; ii < num_of_fuzzy_inputs; ii++)
	{
		inputs_to_fuzzy[ii] = 0.0;
	}
}

/*
DECISION_PROCESS

The DECISION_PROCESS function is called every cycle as long as there is a non-zero command.
It is expected to call a command function based on commandInData->type.

*/
void FUZZY_MODULE::DECISION_PROCESS()
{
	switch(commandInData->type)
	{
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	case FUZZY_INIT_TYPE:
		INIT((FUZZY_INIT *)commandInData);
		break;

	case FUZZY_HALT_TYPE:
		HALT((FUZZY_HALT *)commandInData);
		break;

	case FUZZY_RUN_TYPE:
		RUN((FUZZY_RUN *)commandInData);
		break;

	case FUZZY_SET_DEFUZZ_TYPE:
		SET_DEFUZZ((FUZZY_SET_DEFUZZ *)commandInData);
		break;

	case FUZZY_SET_INPUT_GAINS_TYPE:
		SET_INPUT_GAINS((FUZZY_SET_INPUT_GAINS *)commandInData);
		break;

	case FUZZY_SET_IN_MF_TYPE:
		SET_IN_MF((FUZZY_SET_IN_MF *)commandInData);
		break;

	case FUZZY_SET_IN_UD_TYPE:
		SET_IN_UD((FUZZY_SET_IN_UD *)commandInData);
		break;

	case FUZZY_SET_OUTPUT_GAINS_TYPE:
		SET_OUTPUT_GAINS((FUZZY_SET_OUTPUT_GAINS *)commandInData);
		break;

	case FUZZY_SET_OUT_MF_TYPE:
		SET_OUT_MF((FUZZY_SET_OUT_MF *)commandInData);
		break;

	case FUZZY_SET_OUT_UD_TYPE:
		SET_OUT_UD((FUZZY_SET_OUT_UD *)commandInData);
		break;

	case FUZZY_SET_PREMISE_TYPE:
		SET_PREMISE((FUZZY_SET_PREMISE *)commandInData);
		break;
		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	default:
		logError("The command %d is not recognized.",commandInData->type);
		break;
	}
}

/*
POST_PROCESS

The POST_PROCESS function is called every cycle after DECISION_PROCESS is called
but before the status and the subordinates commands  have been written.
It is intended to be used for tasks such as output filters that should
be performed every cycle regardless of the current command or state.

*/
void FUZZY_MODULE::POST_PROCESS()
{
	static int input_ud_index = 0, output_ud_index = 0;
	static int input_mf_index = 0, output_mf_index = 0;
	int mf_index;
	// Post-Processing Code
	memcpy(fuzzy_status->inputs_to_fuzzy, inputs_to_fuzzy, 
		num_of_fuzzy_inputs*sizeof(double));
	memcpy(fuzzy_status->outputs_of_fuzzy, outputs_of_fuzzy, 
		num_of_fuzzy_outputs*sizeof(double));
	fuzzy_status->defuzz_type = fuzzy_sys->get_defuzz_type();
	fuzzy_status->premise_type = fuzzy_sys->get_premise_type();

	// Input Report

	// If done with this UD switch to the next one
	if (input_mf_index == fuzzy_sys->in_ud[input_mf_index]->get_num_of_mfs())
	{
		input_mf_index = 0;  
		input_ud_index++;
	}

	// If done with all UD start from first
	if (input_ud_index == num_of_fuzzy_inputs)
	{
		input_ud_index = 0;
	}

	// Each Cycle time report status of the current MF in the current UD
	fuzzy_status->num_of_current_in_ud = input_ud_index;
	fuzzy_status->gain_for_in_ud = fuzzy_sys->in_ud[input_ud_index]->ud_gain;

	fuzzy_status->num_of_current_in_mf = input_mf_index;
	fuzzy_status->type_of_current_in_mf = 
	fuzzy_sys->in_ud[input_ud_index]->mfs[input_mf_index].get_type();
	fuzzy_status->center_of_current_in_mf = 
	fuzzy_sys->in_ud[input_ud_index]->mfs[input_mf_index].get_center();
	fuzzy_status->spread_of_current_in_mf = 
	fuzzy_sys->in_ud[input_ud_index]->mfs[input_mf_index].get_spread();

	// Output Report

	// If done with this UD switch to the next one
	if (output_mf_index == fuzzy_sys->out_ud[output_mf_index]->get_num_of_mfs())
	{
		output_mf_index = 0;  
		output_ud_index++;
	}

	// If done with all UD start from first
	if (output_ud_index == num_of_fuzzy_outputs)
	{
		output_ud_index = 0;
	}

	// Each Cycle time report status of the current MF in the current UD
	fuzzy_status->num_of_current_out_ud = output_ud_index;
	fuzzy_status->gain_for_out_ud = fuzzy_sys->out_ud[output_ud_index]->ud_gain;

	fuzzy_status->num_of_current_out_mf = output_mf_index;
	fuzzy_status->type_of_current_out_mf = 
	fuzzy_sys->out_ud[output_ud_index]->mfs[output_mf_index].get_type();
	fuzzy_status->center_of_current_out_mf = 
	fuzzy_sys->out_ud[output_ud_index]->mfs[output_mf_index].get_center();
	fuzzy_status->spread_of_current_out_mf = 
	fuzzy_sys->out_ud[output_ud_index]->mfs[output_mf_index].get_spread();
}

// Command Functions

/*
INIT

Parameter(s):
FUZZY_INIT *cmd_in -- NML Message sent from superior.

Most Modules will have an INIT command.
The INIT function is expected to initialize any variables that may be
in an uninitialized or unknown state, send INIT commands to its subordinates,
wait for the subordinates to be DONE and then inform its superior that it is done.
The state tables should use the STATE_MATCH macro so the diagnostics tool can 
highlight the current line in the state table.

*/
void FUZZY_MODULE::INIT(FUZZY_INIT *cmd_in)
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	if(STATE_MATCH(NEW_COMMAND))
	{
		// Send an INIT command to all subordinates.
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


		stateNext(S1);
		// Reinitialize variables here.

	}
	// Wait for all subordinates to report done.
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	else if(STATE_MATCH(S1,
		1))
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

	{
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
HALT

Parameter(s):
FUZZY_HALT *cmd_in -- NML Message sent from superior.

Most Modules will have an HALT command.
The HALT function is expected to stop any motion or command execution,
send HALT commands to its subordinates,
wait for the subordinates to be DONE and then inform its superior that it is done.
The state tables should use the STATE_MATCH macro so the diagnostics tool can 
highlight the current line in the state table.

*/
void FUZZY_MODULE::HALT(FUZZY_HALT *cmd_in)
{
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.


	if(STATE_MATCH(NEW_COMMAND))
	{
		//Send a HALT command to all subordinates.
		// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

		// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

		stateNext(S1);
	}
	// Wait for all subordinates to report done.
	// RCS-Design-MERGE-DISABLE Edits to the following area will NOT be preserved by the RCS-Design tool.

	else if(STATE_MATCH(S1,
		1))
	// RCS-Design-MERGE-ENABLE Edits after this line will be preserved by the RCS-Design tool.

	{
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}


/*
RUN

Parameter(s):
FUZZY_RUN *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::RUN(FUZZY_RUN *cmd_in)
{
	// Put state table for FUZZY_RUN here.
  if(STATE_MATCH(NEW_COMMAND))
    {
      status = RCS_EXEC;
      stateNext(S1);
    }
  else if(STATE_MATCH(S1))
    {
      fuzzy_sys->control(inputs_to_fuzzy, outputs_of_fuzzy);
    }
  else if(STATE_MATCH(S2))
    {
      // Idle State
    }
}

/*
SET_DEFUZZ

Parameter(s):
FUZZY_SET_DEFUZZ *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_DEFUZZ(FUZZY_SET_DEFUZZ *cmd_in)
{
	// Put state table for FUZZY_SET_DEFUZZ here.
	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		fuzzy_sys->set_defuzz_type(cmd_in->defuzz_type);
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_IN_MF

Parameter(s):
FUZZY_SET_IN_MF *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_IN_MF(FUZZY_SET_IN_MF *cmd_in)
{
	// Put state table for FUZZY_SET_IN_MF here.
  int ii, jj;

	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		ii = cmd_in->num_of_current_in_ud;
		jj = cmd_in->num_of_current_in_mf;

		fuzzy_sys->in_ud[ii]->mfs[jj].set_type(cmd_in->mf_type);
		fuzzy_sys->in_ud[ii]->mfs[jj].set_center(cmd_in->mf_center);
		fuzzy_sys->in_ud[ii]->mfs[jj].set_spread(cmd_in->mf_spread);

		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_INPUT_GAINS

Parameter(s):
FUZZY_SET_INPUT_GAINS *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_INPUT_GAINS(FUZZY_SET_INPUT_GAINS *cmd_in)
{
	// Put state table for FUZZY_SET_INPUT_GAINS here.
	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		fuzzy_sys->set_input_gains(cmd_in->input_gains);
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_IN_UD

Parameter(s):
FUZZY_SET_IN_UD *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_IN_UD(FUZZY_SET_IN_UD *cmd_in)
{
	// Put state table for FUZZY_SET_IN_UD here.
	int ii;

	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		ii = cmd_in->num_of_current_in_ud;

		fuzzy_sys->in_ud[ii]->ud_gain = cmd_in->gain;
		fuzzy_sys->in_ud[ii]->init_ud(cmd_in->first_center, 
			cmd_in->last_center, 
			cmd_in->mf_type, 
			cmd_in->spread);

		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_OUT_MF

Parameter(s):
FUZZY_SET_OUT_MF *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_OUT_MF(FUZZY_SET_OUT_MF *cmd_in)
{
	// Put state table for FUZZY_SET_OUT_MF here.
  int ii, jj;

	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		ii = cmd_in->num_of_current_out_ud;
		jj = cmd_in->num_of_current_out_mf;

		fuzzy_sys->out_ud[ii]->mfs[jj].set_type(cmd_in->mf_type);
		fuzzy_sys->out_ud[ii]->mfs[jj].set_center(cmd_in->mf_center);
		fuzzy_sys->out_ud[ii]->mfs[jj].set_spread(cmd_in->mf_spread);

		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_OUTPUT_GAINS

Parameter(s):
FUZZY_SET_OUTPUT_GAINS *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_OUTPUT_GAINS(FUZZY_SET_OUTPUT_GAINS *cmd_in)
{
	// Put state table for FUZZY_SET_OUTPUT_GAINS here.
	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		fuzzy_sys->set_output_gains(cmd_in->output_gains);
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_OUT_UD

Parameter(s):
FUZZY_SET_OUT_UD *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_OUT_UD(FUZZY_SET_OUT_UD *cmd_in)
{
	// Put state table for FUZZY_SET_OUT_UD here.
	int ii;

	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		ii = cmd_in->num_of_current_out_ud;

		fuzzy_sys->out_ud[ii]->ud_gain = cmd_in->gain;
		fuzzy_sys->out_ud[ii]->init_ud(cmd_in->first_center, 
			cmd_in->last_center, 
			cmd_in->mf_type, 
			cmd_in->spread);

		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

/*
SET_PREMISE

Parameter(s):
FUZZY_SET_PREMISE *cmd_in -- NML Message sent from superior.

*/
void FUZZY_MODULE::SET_PREMISE(FUZZY_SET_PREMISE *cmd_in)
{
	// Put state table for FUZZY_SET_PREMISE here.
	if(STATE_MATCH(NEW_COMMAND))
	{
		status = RCS_EXEC;
		stateNext(S1);
	}
	else if(STATE_MATCH(S1))
	{
		fuzzy_sys->set_premise_type(cmd_in->premise_type);
		status = RCS_DONE;
		stateNext(S2);
	}
	else if(STATE_MATCH(S2))
	{
		// Idle State
	}
}

